@@ -9,16 +9,36 @@
  */
 angular.module('groongaAdminApp')
   .controller('TableSearchController', [
-    '$scope', '$routeParams', '$location', '$q', '$http', '$filter',
-    function ($scope, $routeParams, $location, $q, $http, $filter) {
+    '$scope', '$routeParams', '$location', '$http', '$filter', 'schemaLoader',
+    function ($scope, $routeParams, $location, $http, $filter, schemaLoader) {
+      var schema;
       var client = new GroongaClient($http);
 
+      function findElement(array, finder) {
+        var i, length;
+        length = array.length;
+        for (i = 0; i < length; i++) {
+          var element = array[i];
+          if (finder(element)) {
+            return element;
+          }
+        }
+        return undefined;
+      }
+
       function computeCurrentPage(offset) {
         return Math.ceil((parseInt(offset) + 1) / $scope.nRecordsInPage);
       }
 
       function initialize() {
-        $scope.table = $routeParams.table;
+        $scope.orderedTimeColumnUnits = TimeUnit.getOrderedUnits();
+
+        $scope.table = {
+          name: $routeParams.table,
+          allColumns: [],
+          timeColumns: [],
+          indexedColumns: []
+        };
         $scope.style = 'table';
         $scope.response = {
           rawData: [],
@@ -28,16 +48,16 @@ angular.module('groongaAdminApp')
           elapsedTimeInMilliseconds: 0,
           nTotalRecords: 0
         };
-        $scope.allTables = [];
-        $scope.allColumns = [];
-        $scope.timeColumns = [];
-        $scope.indexedColumns = [];
         $scope.commandLine = '';
         $scope.message = '';
-        $scope.currentPage = 1;
-        $scope.nRecordsInPage = 10;
-        $scope.maxNPages = 10;
         $scope.parameters = angular.copy($location.search());
+        if ($scope.parameters.limit && $scope.parameters.limit > 0) {
+          $scope.nRecordsInPage = $scope.parameters.limit;
+        } else {
+          $scope.nRecordsInPage = 10;
+        }
+        $scope.currentPage = computeCurrentPage($scope.parameters.offset || 0);
+        $scope.maxNPages = 10;
 
         $scope.search = search;
         $scope.incrementalSearch = incrementalSearch;
@@ -53,6 +73,13 @@ angular.module('groongaAdminApp')
         return names.join(',');
       }
 
+      function packMatchColumns(columns) {
+        var names = columns.map(function(column) {
+          return column.indexName;
+        });
+        return names.join(',');
+      }
+
       function packSortColumns(columns) {
         var keys = columns.map(function(column) {
           if (column.sort === 'ascending') {
@@ -89,7 +116,10 @@ angular.module('groongaAdminApp')
       }
 
       function buildFilter() {
-        var timeQueries = $scope.timeColumns.filter(function(column) {
+        $scope.table.timeColumns.forEach(function(column) {
+          column.syncFromRange();
+        });
+        var timeQueries = $scope.table.timeColumns.filter(function(column) {
           return column.start || column.end;
         }).map(function(column) {
           var operator;
@@ -125,24 +155,25 @@ angular.module('groongaAdminApp')
       function buildParameters() {
         var parameters = angular.copy($scope.parameters);
 
-        var matchColumns = $scope.indexedColumns.filter(function(column) {
+        var matchColumns = $scope.table.indexedColumns.filter(function(column) {
           return column.inUse;
         });
-        parameters.match_columns = packColumns(matchColumns);
+        parameters.match_columns = packMatchColumns(matchColumns);
 
-        var outputColumns = $scope.allColumns.filter(function(column) {
+        var outputColumns = $scope.table.allColumns.filter(function(column) {
           return column.output;
         });
         parameters.output_columns = packColumns(outputColumns);
 
+        parameters.offset = ($scope.currentPage - 1) * $scope.nRecordsInPage;
         parameters.limit = $scope.nRecordsInPage;
 
-        var sortColumns = $scope.allColumns.filter(function(column) {
+        var sortColumns = $scope.table.allColumns.filter(function(column) {
           return column.sort;
         });
         parameters.sortby = packSortColumns(sortColumns);
 
-        var drilldowns = $scope.allColumns.filter(function(column) {
+        var drilldowns = $scope.table.allColumns.filter(function(column) {
           return column.drilldown;
         });
         parameters.drilldown = packColumns(drilldowns);
@@ -155,9 +186,7 @@ angular.module('groongaAdminApp')
       }
 
       function search() {
-        var parameters = buildParameters();
-        parameters.offset = ($scope.currentPage - 1) * $scope.nRecordsInPage;
-        $location.search(parameters);
+        $location.search(buildParameters());
       }
 
       function incrementalSearch() {
@@ -183,7 +212,7 @@ angular.module('groongaAdminApp')
       }
 
       function toggleSort(column) {
-        var columnInfo = $scope.allColumns.find(function(columnInfo) {
+        var columnInfo = findElement($scope.table.allColumns, function(columnInfo) {
           return columnInfo.name === column.name;
         });
         if (!columnInfo) {
@@ -207,30 +236,13 @@ angular.module('groongaAdminApp')
         incrementalSearch();
       }
 
-      function isTableType(type) {
-        return $scope.allTables.some(function(table) {
-          return table.name === type;
-        });
-      }
-
-      function isTextType(type) {
-        switch (type) {
-        case 'ShortText':
-        case 'Text':
-        case 'LongText':
-          return true;
-        default:
-          return false;
-        }
-      }
-
       function selectDrilldown(key, value) {
         var queryKey = key;
-        var column = $scope.allColumns.find(function(column) {
+        var column = findElement($scope.table.allColumns, function(column) {
           return column.name === key;
         });
         if (column) {
-          if (isTableType(column.type)) {
+          if (column.valueType.isReferenceType) {
             queryKey += '._key';
           }
         }
@@ -287,22 +299,110 @@ angular.module('groongaAdminApp')
           type: column.range,
           output: output,
           drilldown: drilldown,
-          sort: sort
+          sort: sort,
+          indexes: column.indexes || [],
+          valueType: column.valueType
         };
       }
 
-      function addColumn(columnInfo) {
-        $scope.allColumns.push(columnInfo);
-        if (columnInfo.type === 'Time') {
-          var timeColumnInfo = {
-            name: columnInfo.name,
-            start: null,
-            startIncluded: true,
-            end: null,
-            endIncluded: true
-          };
-          $scope.timeColumns.push(timeColumnInfo);
+      function addTimeColumn(columnInfo) {
+        if (columnInfo.type !== 'Time') {
+          return;
         }
+
+        var timeColumnInfo = {
+          name: columnInfo.name,
+          start: null,
+          startIncluded: true,
+          end: null,
+          endIncluded: true,
+          unit: TimeUnit.units.day,
+          range: [0, 0],
+          syncFromRange: function() {
+            if (this.range[0] === 0 && this.range[1] === 0) {
+              return;
+            }
+            this.start = this.unit.percentToDate(this.range[0] / 100);
+            this.end = this.unit.percentToDate(this.range[1] / 100);
+          },
+          syncToRange: function() {
+            this.unit = TimeUnit.findByDateRange(this.start, this.end);
+            this.updateRangeByDateRange(this.start, this.end);
+          },
+          updateRangeByDateRange: function(start, end) {
+            if (start && end) {
+              this.range = [
+                this.unit.dateToPercent(start) * 100,
+                this.unit.dateToPercent(end) * 100
+              ];
+            } else if (start) {
+              this.range = [
+                this.unit.dateToPercent(start) * 100,
+                100
+              ];
+            } else if (end) {
+              this.range = [
+                0,
+                this.unit.dateToPercent(end) * 100
+              ];
+            } else {
+              this.range = [0, 0];
+            }
+          },
+          formater: function(value) {
+            var date = timeColumnInfo.unit.percentToDate(value / 100);
+            return date.toLocaleString();
+          },
+          onRangeChange: function() {
+            timeColumnInfo.syncFromRange();
+            incrementalSearch();
+          },
+          onUnitChange: function() {
+            timeColumnInfo.updateRangeByDateRange(timeColumnInfo.start,
+                                                  timeColumnInfo.end);
+          }
+        };
+        $scope.table.timeColumns.push(timeColumnInfo);
+      }
+
+      function addIndexedColumn(columnInfo) {
+        if (columnInfo.indexes.length === 0) {
+          return;
+        }
+
+        var isTextType = false;
+        if (columnInfo.valueType.isReferenceType) {
+          var table = schema.tables[columnInfo.valueType.name];
+          isTextType = table.keyType.isTextType;
+        } else {
+          isTextType = columnInfo.valueType.isTextType;
+        }
+
+        if (!isTextType) {
+          return;
+        }
+
+        var matchColumns = $scope.parameters.match_columns;
+        var indexName = columnInfo.name;
+        if (columnInfo.valueType.isReferenceType) {
+          indexName += '._key';
+        }
+        var inUse = true;
+        if (matchColumns) {
+          inUse = (matchColumns.indexOf(indexName) !== -1);
+        }
+        var indexedColumnInfo = {
+          name: columnInfo.name,
+          indexName: indexName,
+          inUse: inUse
+        };
+        $scope.table.indexedColumns.push(indexedColumnInfo);
+      }
+
+      function addColumn(columnInfo) {
+        $scope.table.allColumns.push(columnInfo);
+        addTimeColumn(columnInfo);
+        addIndexedColumn(columnInfo);
       }
 
       function applyTimeQueries() {
@@ -335,7 +435,7 @@ angular.module('groongaAdminApp')
             var startBorder = parts[3];
             var end = parts[4];
             var endBorder = parts[5];
-            timeColumn = $scope.timeColumns.find(function(column) {
+            timeColumn = findElement($scope.table.timeColumns, function(column) {
               return column.name === columnName;
             });
             if (!timeColumn) {
@@ -345,12 +445,13 @@ angular.module('groongaAdminApp')
             timeColumn.startBorder = fromBetweenBorder(startBorder);
             timeColumn.end = fromGroongaTime(end);
             timeColumn.endBorder = fromBetweenBorder(endBorder);
+            timeColumn.syncToRange();
           } else if (/(<=|<|>|=>)/.test(condition)) {
             parts = condition.split(/(<=|<|>|=>)/);
             columnName = parts[0];
             operator = parts[1];
             time = parts[2];
-            timeColumn = $scope.timeColumns.find(function(column) {
+            timeColumn = findElement($scope.table.timeColumns, function(column) {
               return column.name === columnName;
             });
             if (!timeColumn) {
@@ -374,116 +475,14 @@ angular.module('groongaAdminApp')
               timeColumn.startBorder = 'include';
               break;
             }
+            timeColumn.syncToRange();
           }
         });
       }
 
-      function extractColumnsInfo(table, columns) {
-        columns.forEach(function(column) {
-          if (!column.isIndex) {
-            return;
-          }
-          if (column.range !== $scope.table) {
-            return;
-          }
-          var matchColumns = $scope.parameters.match_columns;
-          column.sources.forEach(function(source) {
-            var localName;
-            if (source.indexOf('.') === -1) {
-              localName = '_key';
-            } else {
-              localName = source.split('.')[1];
-            }
-
-            var indexName = localName;
-            var sourceColumn = $scope.allColumns.find(function(column) {
-              return column.name === localName;
-            });
-            if (sourceColumn) {
-              var targetType = sourceColumn.type;
-              var isTableTypeSource = isTableType(targetType);
-              if (isTableTypeSource) {
-                var table = $scope.allTables.find(function(table) {
-                  return table.name === targetType;
-                });
-                targetType = table.domain;
-              }
-              if (!isTextType(targetType)) {
-                return;
-              }
-
-              if (isTableTypeSource) {
-                indexName += '._key';
-              }
-            }
-
-            var inUse = true;
-            if (matchColumns) {
-              inUse = matchColumns.indexOf(indexName) !== -1;
-            }
-
-            $scope.indexedColumns.push({
-              name: indexName,
-              label: localName,
-              inUse: inUse
-            });
-          });
-        });
-      }
-
-      function extractTableInfo(table) {
-        return client.execute('column_list', {table: table.name})
-          .success(function(response) {
-            extractColumnsInfo(table, response.columns());
-          });
-      }
-
-      function fillOptions() {
-        client.execute('table_list')
-          .success(function(response) {
-            $scope.allTables = response.tables();
-
-            var idColumn = {
-              name: '_id',
-              range: 'UInt32'
-            };
-            addColumn(createColumnInfo(idColumn));
-
-            client.execute('column_list', {table: $scope.table})
-              .success(function(response) {
-                var columns = response.columns();
-
-                columns.forEach(function(column) {
-                  if (column.isIndex) {
-                    return;
-                  }
-                  addColumn(createColumnInfo(column));
-                });
-                applyTimeQueries();
-
-                var currentTable = $scope.allTables.find(function(table) {
-                  return table.name === $scope.table;
-                });
-                extractColumnsInfo(currentTable, columns);
-
-                var tasks = [];
-                $scope.allTables.forEach(function(table) {
-                  if (table.name === $scope.table) {
-                    return;
-                  }
-                  tasks.push(extractTableInfo(table));
-                });
-                $q.all(tasks)
-                  .then(function() {
-                    select(buildParameters());
-                  });
-              });
-          });
-      }
-
       function select(userParameters) {
         var parameters = {
-          table: $scope.table
+          table: $scope.table.name
         };
         angular.forEach(userParameters, function(value, key) {
           if (key in parameters) {
@@ -508,7 +507,7 @@ angular.module('groongaAdminApp')
           $scope.response.nTotalRecords = response.nTotalRecords();
           $scope.response.columns = response.columns();
           $scope.response.columns.forEach(function(column) {
-            var columnInfo = $scope.allColumns.find(function(columnInfo) {
+            var columnInfo = findElement($scope.table.allColumns, function(columnInfo) {
               return columnInfo.name === column.name;
             });
             if (columnInfo) {
@@ -545,5 +544,19 @@ angular.module('groongaAdminApp')
       }
 
       initialize();
-      fillOptions();
+      schemaLoader()
+        .then(function(_schema) {
+          schema = _schema;
+          var table = schema.tables[$scope.table.name];
+          angular.forEach(table.columns, function(column) {
+            addColumn(createColumnInfo(column));
+          });
+          applyTimeQueries();
+
+          var parameters = buildParameters();
+          if ($scope.parameters.offset) {
+            parameters.offset = $scope.parameters.offset;
+          }
+          select(parameters);
+        });
     }]);
